from updatorr.handler_base import GenericPrivateTrackerHandler
from updatorr.utils import register_tracker_handler
import urllib2


class AnidubHandler(GenericPrivateTrackerHandler):
    """This class implements .torrent files downloads
    for http://tr.anidub.com tracker."""

    login_url = 'http://tr.anidub.com/takelogin.php'
    cookie_logged_in = 'uid'

    def get_download_link(self):
        """Tries to find .torrent file download link at forum thread page
        and return that one."""
        response, page_html = self.get_resource(self.resource_url)
        page_links = self.find_links(page_html)
        download_link = None
        for page_link in page_links:
            if 'login.php?returnto=' in page_link:
                download_link = None
                self.debug('Login is required to download torrent file.')
                if self.login(self.get_settings('login'), self.get_settings('password')):
                    download_link = self.get_download_link()
            if 'download.php?id=' in page_link:
                download_link = 'http://tr.anidub.com/%s' % urllib2.unquote(page_link).replace('&amp;', '&')
        return download_link

register_tracker_handler('tr.anidub.com', AnidubHandler)



import os
import re
import tempfile
import urllib2

from urllib import urlencode
from updatorr.utils import Cookies


# This regex is used to get all hyperlinks from html.
RE_LINK = re.compile(r'href\s*=\s*"\s*([^"]+)\s*"[^>]*>', re.S | re.I | re.M)

# Cookie registry holds cookies from different tracker hosts.
COOKIES_REGISTRY = {}


class BaseTrackerHandler(object):
    """Base torrent tracker handler class offering
    helper methods for its ancestors."""

    # This tells Updatorr that login procedure is required.
    login_required = True

    # Torrent tracker host this handler is associated with.
    tracker_host = None
    # Torrent data from Deluge session.
    torrent_data = {}
    # Torrent hash from Deluge session data.
    torrent_hash = None
    # Resource URL from torrent comment.
    resource_url = None

    # Updatorr plugin logger instance.
    _log = None
    # Tracker specific settings (e.g. credentials).
    _tracker_settings = {}
    # Occured error description.
    _error_text = ''

    def __init__(self, tracker_host, torrent_data, logger):
        self.tracker_host = tracker_host
        self.torrent_data = torrent_data
        self._log = logger
        self.torrent_hash = torrent_data.get('hash')
        self.resource_url = torrent_data.get('comment')

    def set_settings(self, setting):
        """Stores tracker specific settings (e.g. credentials)
        for further usage."""
        if setting is None:
            setting = {}
        self._tracker_settings = setting
        # Initialize cookies data from previous handler session.
        if 'cookies' in setting:
            self.reset_cookies(setting['cookies'])

    def get_settings(self, field=None):
        """Returns tracker specific settings dictionary.
        Use `field` param to get definite settings value.
        """
        if field is None:
            return self._tracker_settings
        return self._tracker_settings.get(field)

    def login(self, login, password):
        """Tracker login procedure should be implemented
        in child class if necessary."""
        return False

    def get_error_text(self):
        """Returns handling error description if any."""
        return self._error_text

    def dump_error(self, text):
        """Stores handling error description for further usage."""
        self._error_text = text
        self.debug('Error: %s' % text)

    def get_resource(self, url, form_data=None):
        """Returns an HTTP resource data from given URL.
        If a dictionary is passed in `form_data` POST HTTP method
        would be used to pass data to resource (even if that dictionary is empty).

        """
        self.debug('Getting page at %s ...' % url)

        if form_data is not None:
            form_data = urlencode(form_data)

        opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(self.get_cookies()))
        request = urllib2.Request(url, form_data,
                                  {'User-agent': 'Mozilla/5.0 (Ubuntu; X11; Linux i686; rv:8.0) Gecko/20100'})

        try:
            response = opener.open(request)
        except urllib2.URLError:
            return {}, ''

        return response.info(), response.read()

    def store_tmp_torrent(self, file_contents):
        """Stores downloaded .torrent file contents
        in a temporary file within a filesystem.
        Returns a filepath to that file.

        """
        fd, fpath = tempfile.mkstemp()
        f = os.fdopen(fd, 'w')
        f.write(file_contents)
        f.close()
        return fpath

    def set_cookie(self, name, value):
        """Stores a single cookie into cookies registry
        for further usage.

        """
        self.get_cookies().add(name, value)

    def reset_cookies(self, cookies_dict):
        """Initializes cookies registry for current tracker
        with data from the given dictionary.

        """
        global COOKIES_REGISTRY
        COOKIES_REGISTRY[self.tracker_host] = Cookies(init_from_dict=cookies_dict)

    def get_cookies(self, as_dict=False):
        """Returns cookies for this tracker.
        If ``as_dict`` is True returns a dictionary with cookies data,
        otherwise return CookieJar object.

        """
        if self.tracker_host not in COOKIES_REGISTRY:
            COOKIES_REGISTRY[self.tracker_host] = Cookies()
        if as_dict:
            return COOKIES_REGISTRY[self.tracker_host].to_dict()
        return COOKIES_REGISTRY[self.tracker_host]

    def find_links(self, page_html):
        """Returns a list witj hyperlinks found in supplied html."""
        links = []
        for match in re.finditer(RE_LINK, page_html):
            links.append(match.group(1))
        return links

    def get_torrent_file(self):
        """This method should be implemented in torrent tracker
        handler class and must return .torrent filepath
        on success or None on failure."""
        return None

    def debug(self, text):
        """A shortcut to store debug info."""
        self._log.debug(text)


class GenericTrackerHandler(BaseTrackerHandler):
    """Generic torrent tracker handler class implementing
    most common tracker handling methods."""

    def get_id_from_link(self):
        """Returns forum thread identifier from full thread URL."""
        return self.resource_url.split('=')[1]

    def get_torrent_file(self):
        """This is the main method which returns
        a filepath to the downloaded file."""
        torrent_file = None
        download_link = self.get_download_link()
        if download_link is None:
            self.dump_error('Cannot find torrent file download link at %s' % self.resource_url)
        else:
            self.debug('Torrent download link found: %s' % download_link)
            torrent_file = self.download_torrent(download_link)
        return torrent_file

    def get_download_link(self):
        """Tries to find .torrent file download link at forum thread page
        and return that one."""
        raise NotImplementedError()

    def download_torrent(self, url):
        """Gets .torrent file contents from given URL and
        stores that in a temporary file within a filesystem.
        Returns a path to that file.

        """
        raise NotImplementedError()


class GenericPublicTrackerHandler(GenericTrackerHandler):
    """Generic torrent tracker handler class implementing
    most common handling methods for public trackers."""

    login_required = False

    def get_id_from_link(self):
        """Returns forum thread identifier from full thread URL."""
        return self.resource_url.split('/')[-1]

    def download_torrent(self, url):
        """Gets .torrent file contents from given URL and
        stores that in a temporary file within a filesystem.
        Returns a path to that file.

        """
        self.debug('Downloading torrent file from %s ...' % url)
        # That was a check that user himself visited torrent's page ;)
        response, contents = self.get_resource(url)
        return self.store_tmp_torrent(contents)


class GenericPrivateTrackerHandler(GenericPublicTrackerHandler):
    """Generic torrent tracker handler class implementing
    most common handling methods for private trackers (that require
    registration).

    """

    login_required = True
    login_url = None

    # Cookie to verify that a log in was successful.
    cookie_logged_in = None
    logged_in = False
    # Stores a number of login attempts to prevent recursion.
    login_counter = 0

    def get_login_form_data(self, login, password):
        """Should return a dictionary with data to be pushed
        to authorization form.

        """
        return {'username': login, 'password': password}

    def login(self, login, password):
        """Implements tracker login procedure."""
        self.logged_in = False

        if login is None or password is None:
            return False

        self.login_counter += 1

        # No recursion wanted.
        if self.login_counter > 1:
            return False

        self.debug('Trying to login at %s ...' % self.login_url)
        self.get_resource(self.login_url, self.get_login_form_data(login, password))
        cookies = self.get_cookies()

        # Login success check.
        if cookies.get(self.cookie_logged_in) is not None:
            self.logged_in = True
        return self.logged_in

    def before_download(self):
        """Used to perform some required actions right before .torrent download.
        E.g.: to set a sentinel cookie that allows the download."""
        return True

    def download_torrent(self, url):
        """Gets .torrent file contents from given URL and
        stores that in a temporary file within a filesystem.
        Returns a path to that file.

        """
        self.debug('Downloading torrent file from %s ...' % url)
        self.get_cookies()
        self.before_download()
        contents = self.get_resource(url, {})[1]
        return self.store_tmp_torrent(contents)



import os
import logging
import threading
import base64
import pkgutil
import time
import re

import deluge.configmanager
import deluge.component as component
from deluge.common import is_url
from deluge.event import DelugeEvent
from deluge.core.rpcserver import export
from deluge.plugins.pluginbase import CorePluginBase
from twisted.internet.task import LoopingCall

# Line below is required to import tracker handler on fly.
import updatorr.tracker_handlers
from updatorr.utils import *


log = logging.getLogger(__name__)

# Import tracker handlers on fly.
# It is an .egg-friendly alternative to os.listdir() walking.
for mloader, pname, ispkg in pkgutil.iter_modules(updatorr.tracker_handlers.__path__):
    log.info('Updatorr Importing tracker handler file %s' % pname)
    __import__('updatorr.tracker_handlers.%s' % pname)


# Default plugin preferences.
DEFAULT_PREFS = {
    'last_walk': 0,
    'walk_period': 24,
    'trackers_settings': {},
    'torrents_to_update': []  # That might have been set(), except config serialization.
}

# This regex is used to get hyperlink from torrent comment.
RE_LINK = re.compile(r'(?P<url>https?://[^\s]+)')


class UpdatorrUpdateDoneEvent(DelugeEvent):
    """This event fires up when a torrent is updated."""
    def __init__(self, torrent_id):
        self._args = [torrent_id]


class UpdatorrErrorEvent(DelugeEvent):
    """This event fires up when an error occures in tracker handler."""
    def __init__(self, torrent_id, error_text):
        self._args = [torrent_id, error_text]


class UpdatorrUpdatesCheckStartedEvent(DelugeEvent):
    """This event fires up when torrent updates check is started."""
    pass


class UpdatorrUpdatesCheckFinishedEvent(DelugeEvent):
    """This event fires up when torrent updates check is finished."""
    pass


class Core(CorePluginBase):

    walking = False
    plugin_id = 'Updatorr'

    def enable(self):
        """This one fires when plugin is enabled."""
        self.plugin = component.get('CorePluginManager')
        self.plugin.register_status_field(self.plugin_id, self.get_status_label)

        self.core = component.get('Core')
        self.torrents = self.core.torrentmanager.torrents

        self.config = deluge.configmanager.ConfigManager('updatorr.conf', DEFAULT_PREFS)
        self.torrents_to_update = self.config['torrents_to_update']
        self.walk_period = self.config['walk_period']
        self.last_walk = self.config['last_walk']
        self.trackers_settings = self.config['trackers_settings']

        self.update_trackers_settings()

        self.filter_manager = component.get('FilterManager')
        self.filter_manager.register_tree_field(self.plugin_id, self.get_filters_initial)

        # We will check whether it's time to go for updates every 60 seconds.
        self.walk_torrents_timer = LoopingCall(self.run_walker)
        self.walk_torrents_timer.start(60)


    def disable(self):
        """That one fires when plugin is disabled."""
        self.walk_torrents_timer.stop()
        self.filter_manager.deregister_tree_field(self.plugin_id)
        self.plugin.deregister_status_field(self.plugin_id)
        self.save_config()

    def update(self):
        """This one fires every second while plugin is enabled."""
        pass

    UPDATE_STATES = {True: 'On', False: 'Off'}

    def get_status_label(self, torrent_id):
        """This one is to update filter tree numbers.
        It is called every time torrent status is changed."""
        return self.UPDATE_STATES[self.check_is_to_update(torrent_id)]

    def get_filters_initial(self):
        """That are initial filter tree values."""
        return {'On': 0, 'Off': 0, 'All': len(self.torrents.keys())}

    def update_trackers_settings(self):
        """Returns registered handlers dictionary."""
        for domain, handler in get_registered_handlers().items():
            if domain not in self.trackers_settings:
                domain_dict = {
                    'login_required': handler.login_required,
                    'login': '',
                    'password': '',
                    'cookies': None
                }
                self.trackers_settings[domain] = domain_dict
            else:
                self.trackers_settings[domain].update({'login_required': handler.login_required})

    @export
    def get_status(self):
        """Returns tuple with Updatorr status data:
        last walk time, walk period in hours, is currently walking."""
        return self.last_walk, self.walk_period, self.walking

    @export
    def test_login(self, domain, login, password):
        """Launches login procedure for tracker domain.
        Returns True on success, overwise - False."""
        handler = get_tracker_handler({'comment': domain}, log)
        handler.set_settings(self.trackers_settings.get(domain))
        if handler is not None:
            return handler.login(login=login, password=password)
        return None

    @export
    def is_walking(self):
        """Returns boolean to identify whether update
        proccess is on the run."""
        return self.walking

    @export
    def run_walker(self, force=False):
        """Runs update process in a separate thread
        if it is a hight time for it and it's not already started."""
        if not force:
            now = time.time()
            next_walk = int(self.last_walk) + (int(self.walk_period) * 3600)
            log.debug('Updatorr run walker: walking=%s; next_walk=%s; now=%s' % (self.walking, next_walk, now))
            if self.walking:
                return False
            if next_walk > now:
                return False
        threading.Thread(target=self.walk, kwargs={'force': force}).start()
        return True

    def walk(self, force=False):
        """Implemets automatic torrent updates process.
        Automatic update is available for torrents selected by user
        and having tracker's page URL in torrent's `comment` field.

        Besides that torrent a tracker handler class should be
        associated with domain from the URL mentioned above.

        If `force` set to a list of torrent IDs, only those
        torrents will be checked for updates.
        If `force` is False every torrent scheduled to updates
        by used will be checked.

        """

        # To prevent possible concurent runs.
        self.walking = True

        log.info('Updatorr walking...')
        component.get('EventManager').emit(UpdatorrUpdatesCheckStartedEvent())

        allow_last_walk_update = False

        if isinstance(force, list):
            torrents_list = force
        else:
            torrents_list = self.torrents_to_update

        for torrent_id in torrents_list:
            try:
                torrent_data = self.core.get_torrent_status(torrent_id, [])
            except KeyError:
                log.debug('Updatorr \tSKIPPED No torrent with id %s listed [yet]' % torrent_id)
                continue
            log.info('Updatorr Processing %s ...' % torrent_data['name'])
            # Remove not url data from comment
            torrent_data['comment'] = RE_LINK.search(torrent_data['comment']).group('url')
            if not is_url(torrent_data['comment']):
                log.info('Updatorr \tSKIPPED No URL found in torrent comment')
                continue
            # From now on we consider that update took its place.
            # If only this update is not forced.
            if not force:
                allow_last_walk_update = True
            tracker_handler = get_tracker_handler(torrent_data, log)
            if tracker_handler is None:
                self.dump_error(torrent_id, 'Unable to find tracker handler for %s' % torrent_data['comment'])
                continue
            tracker_handler.set_settings(self.trackers_settings.get(tracker_handler.tracker_host))
            new_torrent_filepath = tracker_handler.get_torrent_file()
            if new_torrent_filepath is None:
                self.dump_error(torrent_id, 'Error in tracker handling: %s' % tracker_handler.get_error_text())
                continue

            # Let's store cookies form that tracker to enter without logins in future sessions.
            self.trackers_settings[tracker_handler.tracker_host]['cookies'] = tracker_handler.get_cookies(as_dict=True)

            new_torrent_contents = read_torrent_file(new_torrent_filepath)
            new_torrent_info = read_torrent_info(new_torrent_contents)
            if torrent_data['hash'] == new_torrent_info['hash']:
                log.info('Updatorr \tSKIPPED Torrent is up-to-date')
                continue
            log.info('Updatorr \tTorrent update is available')

            new_torrent_prefs = get_new_prefs(torrent_data, new_torrent_info)
            added_torrent_id = self.core.add_torrent_file(None, base64.encodestring(new_torrent_contents), new_torrent_prefs)

            if added_torrent_id is not None:
                self.core.remove_torrent(torrent_id, False)
                log.info('Updatorr \tTorrent is updated')
                # Fire up update finished event.
                component.get('EventManager').emit(UpdatorrUpdateDoneEvent(new_torrent_info['hash']))
                # Add new torrent hash to continue autoupdates.
                self.set_items_to_update(new_torrent_info['hash'], True)
                # Remove old torrent from autoupdates list.
                self.set_items_to_update(torrent_id, False)
            else:
                self.dump_error(torrent_id, 'Unable to replace current torrent with a new one')

            # No littering, remove temporary .torrent file.
            os.remove(new_torrent_filepath)

        if allow_last_walk_update:
            # Remember lastrun time.
            self.last_walk = time.time()

        log.info('Updatorr walk is finished')
        component.get('EventManager').emit(UpdatorrUpdatesCheckFinishedEvent())
        self.walking = False

    def dump_error(self, torrent_id, text):
        """Logs error and fires error event."""
        log.info('Updatorr \tSKIPPED %s' % text)
        component.get('EventManager').emit(UpdatorrErrorEvent(torrent_id, text))

    @export
    def set_items_to_update(self, torrent_id, do_update):
        """Adds or removes given torrent to the `torrents-to-update list`."""
        if do_update:
            if torrent_id not in self.torrents_to_update:
                self.torrents_to_update.append(torrent_id)
        elif torrent_id in self.torrents_to_update:
            self.torrents_to_update.remove(torrent_id)
        self.save_config()

    @export
    def check_is_to_update(self, torrent_id):
        """Checks whether given torrent is set to update. Returns boolean."""
        return torrent_id in self.torrents_to_update

    @export
    def get_items_to_update(self):
        """Retunt a lis of to"""
        return self.torrents_to_update

    @export
    def set_config(self, config=None):
        """Sets the config dictionary of torrent IDs to update."""
        log.debug('Updatorr sets config')
        if config is not None:
            self.walk_period = config['walk_period']
            self.trackers_settings = config['trackers_settings']
        self.save_config()

    @export
    def get_config(self):
        """Returns the config dictionary"""
        log.debug('Updatorr gets config')
        return self.config.config

    def save_config(self):
        """Dumps configuration file to file system ~/.config/deluge/updatorr.conf."""
        # Going through every name to be sure...
        self.update_trackers_settings()
        self.config['walk_period'] = int(self.walk_period)
        self.config['last_walk'] = int(self.last_walk)
        self.config['torrents_to_update'] = self.torrents_to_update
        self.config['trackers_settings'] = self.trackers_settings
        self.config.save()



from updatorr.handler_base import GenericPublicTrackerHandler
from updatorr.utils import register_tracker_handler


class RutorHandler(GenericPublicTrackerHandler):
    """This class implements .torrent files downloads
    for http://rutor.org tracker."""
    
    def get_download_link(self):
        """Tries to find .torrent file download link at forum thread page
        and return that one."""
        linkToFind = 'd.rutor.org/download/%s' % self.get_id_from_link()
        response, page_html = self.get_resource(self.resource_url)
        page_links = self.find_links(page_html)
        download_link = None
        for page_link in page_links:
            if linkToFind in page_link:
                download_link = page_link
                break
        return download_link

register_tracker_handler('rutor.org', RutorHandler)



from deluge.plugins.init import PluginInitBase


VERSION = (0, 1, 7)


class CorePlugin(PluginInitBase):

    def __init__(self, plugin_name):
        from core import Core as _plugin_cls
        self._plugin_cls = _plugin_cls
        super(CorePlugin, self).__init__(plugin_name)


class GtkUIPlugin(PluginInitBase):

    def __init__(self, plugin_name):
        from gtkui import GtkUI as _plugin_cls
        self._plugin_cls = _plugin_cls
        super(GtkUIPlugin, self).__init__(plugin_name)


class WebUIPlugin(PluginInitBase):
    def __init__(self, plugin_name):
        from webui import WebUI as _plugin_cls
        self._plugin_cls = _plugin_cls
        super(WebUIPlugin, self).__init__(plugin_name)



import time
import os

from collections import defaultdict
from cookielib import CookieJar
from copy import deepcopy

from deluge._libtorrent import lt

# Torrent tracker handler classes registry.
TRACKER_HANDLERS = {}


def register_tracker_handler(domain_name, handler_callable):
    """Registers a tracker handler class with some domain name."""
    global TRACKER_HANDLERS
    TRACKER_HANDLERS[domain_name] = handler_callable


def get_registered_handlers(domain=None):
    """Returns registered handlers dictionary."""
    if domain is None:
        return TRACKER_HANDLERS
    return TRACKER_HANDLERS.get(domain)


def get_tracker_handler(torrent_data, logger):
    """Returns an appropriate torrent tracker handler object
    from handlers dictionary basing on resource_url."""
    for host_name in TRACKER_HANDLERS.keys():
        if host_name in torrent_data['comment']:
            return TRACKER_HANDLERS[host_name](host_name, torrent_data, logger)
    return None


def read_torrent_file(filepath):
    """Reads torrent file from filesystem and returns its contents."""
    f = open(filepath, 'rb')
    contents = f.read()
    f.close()
    return contents


def read_torrent_info(file_contents):
    """Returns a dictionary with basic information from torrent
    contents (see `read_torrent_file()`).

    Dict keys:
        hash - Torrent hash.
        files - A list of files within the torrent.

    """
    info_contents = lt.torrent_info(lt.bdecode(file_contents))
    files_from_torrent = [a_file.path for a_file in info_contents.files()]
    info = {'hash': str(info_contents.info_hash()), 'files': files_from_torrent}
    return info


def get_new_prefs(full_prefs, new_torrent_info):
    """Returns a dictionary with preferences for a new torrent session.
    Those preferences are deduces from full preferences of previous
    session of that torrent.

    Those files that were marked as "not to download" in previous session
    whold have the same priority in a new one. Basically all files priorities
    are copied from the previous session.

    """
    new_prefs = deepcopy(full_prefs)

    # These are some items we definitely do not want in a new session.
    info_to_drop = (
        'comment', 'hash', 'peers', 'progress', 'num_seeds', 'ratio', 'total_peers', 'total_wanted',
        'distributed_copies', 'time_added', 'active_time', 'next_announce',
        'tracker', 'tracker_host', 'tracker_status', 'trackers',
        'files', 'file_priorities', 'file_progress', 'num_files',
        'is_seed', 'seed_rank', 'seeding_time', 'download_payload_rate', 'message',
        'num_peers', 'compact', 'total_uploaded', 'total_done', 'num_pieces',
        'total_payload_download', 'total_seeds', 'piece_length', 'all_time_download',
        'name', 'seeds_peers_ratio', 'eta', 'is_finished', 'total_size', 'state', 'upload_payload_rate',
        )

    for pref_name in info_to_drop:
        if pref_name in new_prefs:
            del new_prefs[pref_name]

    new_prefs['download_location'] = new_prefs['save_path']
    new_prefs['mapped_files'] = {}
    new_prefs['file_priorities'] = []

    # Copying files priorities.
    old_priorities = get_files_priorities(full_prefs)
    for a_file in new_torrent_info['files']:
        priority = 1
        if a_file in old_priorities:
            priority = old_priorities[a_file]
        new_prefs['file_priorities'].append(priority)

    # Check for root folder rename
    old_splited = os.path.split(full_prefs['files'][0]['path'])
    if len (old_splited) > 1:
        old_folder = old_splited[0]
        new_folder = os.path.split(new_torrent_info['files'][0])[0]
        if new_folder != old_folder:
            i = 0
            for a_file in new_torrent_info['files']:
                new_prefs['mapped_files'][i] = a_file.replace(new_folder, old_folder)
                i += 1


    return new_prefs


def get_files_priorities(torrent_data):
    """Returns a dictionary with files priorities, where
    filepaths are keys, and priority identifiers are values."""
    files = {}
    walk_counter = 0
    for file in torrent_data['files']:
        files[file['path']] = torrent_data['file_priorities'][walk_counter]
        walk_counter += 1
    return  files


class DummyRequest(object):
    """Fake request object to satisfy CookieJar._cookies_from_attrs_set.
    See ``Cookies`` class.

    """

    def get_full_url(self):
        return ''

    def get_host(self):
        return ''

    def get_header(self, name, default):
        return default


class Cookies(CookieJar):
    """Wrapper around CookieJar with helper methods."""

    def __init__(self, policy=None, init_from_dict=None):
        CookieJar.__init__(self, policy=policy)
        self._now = int(time.time())
        if init_from_dict is not None:
            self.from_dict(init_from_dict)

    def get(self, name, default=None):
        """Returns value for cookie by name.
        If cookie is not found in a jar value of ``default``
        is returned.

        """
        value = default
        for cookie in self:
            if cookie.name==name:
                return cookie.value
        return value

    def add(self, name, value, params=None):
        """Adds cookie into a jar."""
        cookie_tuple = [(name, value)]
        if params is not None:
            cookie_tuple.extend(params)
        cookies = self._cookies_from_attrs_set([cookie_tuple], DummyRequest())
        for  cookie in cookies:
            self.set_cookie(cookie)

    def to_dict(self):
        """Converts CookieJar into a dictionary."""
        output = defaultdict(dict)
        for cookie in self:
            for attr in cookie.__dict__:
                output[cookie.name][attr] = cookie.__dict__[attr]
        return output

    def from_dict(self, source_dict):
        """Populates CookieJar from a dictionary."""
        for name, props in source_dict.items():
            self.add(name, props['value'], props.items())



import os
import logging
import locale
import pkg_resources
from datetime import datetime

import gtk
import deluge.component as component
import deluge.common
from deluge.ui.client import client
from deluge.ui.gtkui import notification
from deluge.plugins.pluginbase import GtkPluginBase
from twisted.internet.defer import Deferred
from twisted.internet.task import LoopingCall


CURRENT_LOCALE = locale.getdefaultlocale()[1]
log = logging.getLogger(__name__)


class UpdatorNotification(notification.Notification):

    _title = 'Deluge Updatorr'
    _message = ''

    def get_torrent_status(self, torrent_id=None):
        if torrent_id is not None:
            component.get('SessionProxy').get_torrent_status(torrent_id,
                ['name', 'num_files', 'total_payload_download']).addCallback(self._on_get_torrent_status)
        else:
            self._on_get_torrent_status({'total_payload_download': 'fake'})

    def _on_get_torrent_status(self, status):
        if status is None:
            return None
        if self.config['ntf_popup']:
            self.popup(status)

    def set_title(self, title):
        self._title = title

    def set_message(self, message):
        self._message = message

    def get_title(self):
        return self._enc(self._title)

    def get_message(self):
        return self._enc(self._message)

    def _enc(self, value):
        return deluge.common.xml_encode(value)

    def popup(self, status):
        if not deluge.common.windows_check():
            try:
                import pynotify
            except Exception:
                pass
            else:
                if not pynotify.init('Deluge'):
                    return None
                message = self.get_message()
                if 'name' in status:
                    message = '%s %s' % (message, self._enc('%s' % status['name']))
                self.note = pynotify.Notification(self.get_title(), message)
                self.note.show()


class GtkUI(GtkPluginBase):

    plugin_id = 'Updatorr'
    trackers_data_model = None
    status_bar_item = None

    def get_resource(self, name):
        """Returns resource from data dir."""
        return pkg_resources.resource_filename('updatorr', os.path.join('data', name))

    def enable(self):
        """Triggers when plugin is enabled."""
        self.glade = gtk.glade.XML(self.get_resource('config.glade'))

        self.tv_trackers = self.glade.get_widget('tv_tracker')
        self.ed_login = self.glade.get_widget('ed_login')
        self.ed_password = self.glade.get_widget('ed_password')

        component.get('Preferences').add_page(self.plugin_id, self.glade.get_widget('prefs_box'))
        pmanager = component.get('PluginManager')
        pmanager.register_hook('on_apply_prefs', self.on_apply_prefs)
        # TODO Cope with tv_tarckers repopulation on plugin reenable.
        pmanager.register_hook('on_show_prefs', self.on_show_prefs)
        self.add_context_menu_items()

        self.status_updator = LoopingCall(self.update_status_bar)
        self.status_updator.start(60)

        client.register_event_handler('UpdatorrUpdateDoneEvent', self.on_update_done_event)
        client.register_event_handler('UpdatorrErrorEvent', self.on_update_error_event)
        client.register_event_handler('UpdatorrUpdatesCheckStartedEvent', self.on_updates_started_event)
        client.register_event_handler('UpdatorrUpdatesCheckFinishedEvent', self.on_updates_finished_event)

    def update_status_bar(self):
        """Updates status bar item."""
        if self.status_bar_item is None:
            self.status_bar_item = component.get('StatusBar').add_item(stock=gtk.STOCK_REFRESH,
                text='', tooltip=_('Updatorr autoupdate status information'))
        client.updatorr.get_status().addCallback(self.get_status)

    DATE_FORMAT = '%b %d %H:%M'

    def get_status(self, status_data):
        """Returns autoupdate status and puts it our status bar item."""
        if status_data[0] == 0:
            self.status_bar_item.set_text(_('Updates are not checked yet'))
        else:
            last_updated = datetime.fromtimestamp(status_data[0]).strftime(self.DATE_FORMAT)
            next_update = datetime.fromtimestamp(int(status_data[0]) + int(status_data[1] * 3600)).strftime(self.DATE_FORMAT)
            # Note that format used is locale-specific and UTF-8 encoding is required for UI output.
            if CURRENT_LOCALE is not None:
                last_updated = last_updated.decode(CURRENT_LOCALE).encode('utf8')
                next_update = next_update.decode(CURRENT_LOCALE).encode('utf8')
            self.status_bar_item.set_text(_('last: %s, next: %s') % (last_updated, next_update))

    def disable(self):
        """Triggers when plugin is disabled."""
        self.status_updator.stop()
        component.get('StatusBar').remove_item(self.status_bar_item)
        self.status_bar_item = None
        self.remove_context_menu_items()
        self.trackers_data_model = None
        component.get('Preferences').remove_page(self.plugin_id)
        component.get('PluginManager').deregister_hook('on_apply_prefs', self.on_apply_prefs)
        component.get('PluginManager').deregister_hook('on_show_prefs', self.on_show_prefs)

    def init_tv_trackers(self):
        """Initializes trackers TreeView in preferences window."""

        # Initialize only once.
        if self.trackers_data_model is None:
            self.trackers_data_model = gtk.ListStore(str, bool, str, str)

            text_cell = gtk.CellRendererText()
            bool_cell = gtk.CellRendererToggle()

            tvc_domain = gtk.TreeViewColumn('Tracker domain', text_cell, text=0)
            tvc_login_req = gtk.TreeViewColumn('Login required', bool_cell)
            tvc_login_req.add_attribute(bool_cell, 'active', 1)
            tvc_login = gtk.TreeViewColumn('Login', text_cell, text=2)
            tvc_password = gtk.TreeViewColumn('Password', text_cell, text=3)
            tvc_password.set_visible(False)

            self.tv_trackers.append_column(tvc_domain)
            self.tv_trackers.append_column(tvc_login_req)
            self.tv_trackers.append_column(tvc_login)
            self.tv_trackers.append_column(tvc_password)

            self.tv_trackers.set_model(self.trackers_data_model)

            self.glade.signal_autoconnect({
                'on_btn_test_clicked': self.on_btn_test_clicked,
                'on_btn_apply_clicked': self.on_btn_apply_clicked,
                'on_tv_tracker_cursor_changed': self.on_tv_tracker_cursor_changed
            })

    def on_tv_tracker_cursor_changed(self, widget):
        """Triggered when trackers TreeView row is selected.
        Puts data from trackers TreeView into credentials Entry controls."""
        list_store, iter = self.tv_trackers.get_selection().get_selected()
        self.tv_current_domain = list_store.get_value(iter, 0)
        self.tv_current_login = list_store.get_value(iter, 2)
        self.tv_current_password = list_store.get_value(iter, 3)

        self.ed_login.set_text(self.tv_current_login)
        self.ed_password.set_text(self.tv_current_password)

    def on_btn_test_clicked(self, widget):
        """Triggers when test credentials button is pushed."""
        data = [self.tv_current_domain, self.ed_login.get_text(), self.ed_password.get_text()]
        client.updatorr.test_login(*data).addCallback(self.show_login_test_result)

    def show_login_test_result(self, success):
        """Shows credentials test result dialog."""
        if success:
            mtype = gtk.MESSAGE_INFO
            message = _('Credentials seem to be correct. Login was successful.')
        else:
            mtype = gtk.MESSAGE_ERROR
            message = _('Unable to login with given credentials.\nNote: some trackers may eventually activate capcha that Updatorr can not handle.')

        dialog = gtk.MessageDialog(flags=gtk.DIALOG_DESTROY_WITH_PARENT, type=mtype, buttons=gtk.BUTTONS_CLOSE, message_format=message)
        dialog.set_title(_('Tracker login test'))

        # Twisted might blow your away to hell eventually.
        result = Deferred()

        def response(dialog, resp):
            dialog.destroy()
            result.callback(resp)
            return False
        dialog.connect('response', response)
        dialog.show_all()
        return result

    def on_btn_apply_clicked(self, widget):
        """Triggers when apply credentials button is pushed.
        Puts data from credentials Entry controls into trackers TreeView.
        """
        # Get selected item for update.
        list_store, iter = self.tv_trackers.get_selection().get_selected()
        list_store.set_value(iter, 2, self.ed_login.get_text())
        list_store.set_value(iter, 3, self.ed_password.get_text())

    def on_update_done_event(self, torrent_id):
        """Triggers when torrent update is successful.
        Notifies user abount the event."""
        n = UpdatorNotification()
        n.set_title(_('Deluge Updatorr: torrent is updated'))
        n.notify(torrent_id)

    def on_update_error_event(self, torrent_id, error_text):
        """Triggers when torrent update failed.
        Notifies user abount the event."""
        n = UpdatorNotification()
        n.set_title(_('Deluge Updatorr Error'))
        n.set_message('%s at ' % error_text)
        n.notify(torrent_id)

    def on_updates_started_event(self):
        """Triggers when torrent update is started.
        Notifies user abount the event."""
        n = UpdatorNotification()
        n.set_message(_('Updates check is started'))
        n.notify(None)

    def on_updates_finished_event(self):
        """Triggers when torrent update is finished.
        Notifies user abount the event."""
        n = UpdatorNotification()
        n.set_message(_('Updates check is finished'))
        n.notify(None)

    # Context menu `toggle autoupdates` item states.
    CONTEXT_UPDATE_CHOICES = {
        True: _('Enable autoupdates'),
        False: _('Disable autoupdates')
    }

    # Context menu `check for updates` item states.
    CONTEXT_RUN_LABELS = {
        True: _('Check updates for selected'),
        False: _('Torrent updates in progress...'),
    }

    # Context menu `check scheduled updates` item states.
    CONTEXT_RUN_ALL_LABELS = {
        True: _('Check updates for scheduled'),
        False: CONTEXT_RUN_LABELS[False],
    }

    def add_context_menu_items(self):
        """Adds Updatorr items into torrents list context menu."""
        menu_bar = component.get('MenuBar')
        self.sep_1 = menu_bar.add_torrentmenu_separator()

        context_menu = menu_bar.torrentmenu
        self.cmenu_item_toggle = gtk.MenuItem(self.CONTEXT_UPDATE_CHOICES[True])
        self.cmenu_item_toggle.connect('activate', self.on_cmenu_item_toggle_activate)

        self.cmenu_item_run = gtk.MenuItem(self.CONTEXT_RUN_LABELS[True])
        self.cmenu_item_run.connect('activate', self.on_cmenu_item_run_activate)

        self.cmenu_item_run_all = gtk.MenuItem(self.CONTEXT_RUN_ALL_LABELS[True])
        self.cmenu_item_run_all.connect('activate', self.on_cmenu_item_run_all_activate)

        context_menu.connect('focus', self.on_cmenu_focus)
        context_menu.append(self.cmenu_item_toggle)
        context_menu.append(self.cmenu_item_run)
        self.sep_2 = menu_bar.add_torrentmenu_separator()
        context_menu.append(self.cmenu_item_run_all)

        self.cmenu_item_toggle.show()
        self.cmenu_item_run.show()
        self.cmenu_item_run_all.show()

    def remove_context_menu_items(self):
        """Removes Updatorr items from torrents list context menu."""
        context_menu = component.get('MenuBar').torrentmenu
        context_menu.remove(self.cmenu_item_toggle)
        context_menu.remove(self.cmenu_item_run)
        context_menu.remove(self.cmenu_item_run_all)
        context_menu.remove(self.sep_1)
        context_menu.remove(self.sep_2)

    def on_cmenu_focus(self, *args, **kwargs):
        """Triggered when torrents list context menu is summoned."""
        client.updatorr.get_items_to_update().addCallback(self.update_cmenu_item_toggle)
        client.updatorr.is_walking().addCallback(self.update_cmenu_item_run)

    def update_cmenu_item_run(self, is_walking):
        """Updates state for check for updates context menu items."""
        self.cmenu_item_run.set_label(self.CONTEXT_RUN_LABELS[not is_walking])
        self.cmenu_item_run_all.set_label(self.CONTEXT_RUN_ALL_LABELS[not is_walking])
        self.cmenu_item_run.set_sensitive(not is_walking)
        self.cmenu_item_run_all.set_sensitive(not is_walking)

    def update_cmenu_item_toggle(self, torrents_to_update):
        """Updates state for `toggle autoupdates` context menu item."""
        states = []
        for torrent_id in self.get_selected_torrents():
            state = torrent_id in torrents_to_update
            states.append(state)
        if states.count(True) >= states.count(False):
            self.cmenu_item_toggle.set_label(self.CONTEXT_UPDATE_CHOICES[False])
        else:
            self.cmenu_item_toggle.set_label(self.CONTEXT_UPDATE_CHOICES[True])

    def on_cmenu_item_toggle_activate(self, widget):
        """Triggered when `toggle autoupdates` context menu item is pushed."""
        enable = False
        if widget.get_label() == self.CONTEXT_UPDATE_CHOICES[True]:
            enable = True
        widget.set_label(self.CONTEXT_UPDATE_CHOICES[not enable])
        for torrent_id in self.get_selected_torrents():
            client.updatorr.set_items_to_update(torrent_id, enable)

    def on_cmenu_item_run_activate(self, widget):
        """Triggered when `check for updates` context menu item is pushed."""
        client.updatorr.run_walker(force=self.get_selected_torrents())

    def on_cmenu_item_run_all_activate(self, widget):
        """Triggered when `check scheduled updates` context menu item is pushed."""
        client.updatorr.run_walker(force=True)

    def get_selected_torrents(self):
        """Returns selected items from torrents list."""
        return component.get('TorrentView').get_selected_torrents()

    def on_apply_prefs(self):
        """Triggers when `apply preferences` button is pushed."""
        trackers_settings = {}

        for row in self.trackers_data_model:
            trackers_settings[row[0]] = {'login': row[2], 'password': row[3]}

        config = {
            'walk_period': self.glade.get_widget('walk_period').get_text(),
            'trackers_settings': trackers_settings
        }
        client.updatorr.set_config(config)

    def on_show_prefs(self):
        """Triggers on plugin preferences window show
        and after `apply preferences` button is pushed."""
        self.init_tv_trackers()
        client.updatorr.get_config().addCallback(self.config_to_ui)

    def config_to_ui(self, config):
        """Reads Updatorr configuration and puts data from it into
        preferences window UI controls."""
        self.glade.get_widget('walk_period').set_value(config['walk_period'])
        self.populate_tv_trackers(config['trackers_settings'])

    def populate_tv_trackers(self, settings):
        """Populates trackers TreeView with trackers data from Updatorr config."""
        self.trackers_data_model.clear()

        for domain, data in settings.items():
            self.trackers_data_model.append([domain, data['login_required'], data['login'], data['password']])

        # Select first tracker row.
        first_el_iter = self.trackers_data_model.get_iter_first()
        self.tv_trackers.get_selection().select_iter(first_el_iter)
        self.tv_trackers.emit('cursor_changed')



from updatorr.handler_base import GenericPrivateTrackerHandler
from updatorr.utils import register_tracker_handler


class RutrackerHandler(GenericPrivateTrackerHandler):
    """This class implements .torrent files downloads
    for http://rutracker.org tracker."""

    login_url = 'http://login.rutracker.org/forum/login.php'
    cookie_logged_in = 'bb_data'

    def get_login_form_data(self, login, password):
        """Returns a dictionary with data to be pushed to authorization form."""
        return {'login_username': login, 'login_password': password, 'login': 'pushed'}

    def before_download(self):
        """Used to perform some required actions right before .torrent download."""
        self.set_cookie('bb_dl', self.get_id_from_link())  # A check that user himself have visited torrent's page ;)

    def get_download_link(self):
        """Tries to find .torrent file download link at forum thread page
        and return that one."""
        response, page_html = self.get_resource(self.resource_url)
        page_links = self.find_links(page_html)
        download_link = None
        for page_link in page_links:
            if 'dl.rutracker.org' in page_link:
                download_link = page_link
                if 'guest' in download_link:
                    download_link = None
                    self.debug('Login is required to download torrent file.')
                    if self.login(self.get_settings('login'), self.get_settings('password')):
                        download_link = self.get_download_link()
                break
        return download_link

# With that one we tell updatetorr to handle links to `rutracker.org` domain with RutrackerHandler class.
register_tracker_handler('rutracker.org', RutrackerHandler)



import os
from setuptools import setup
from updatorr import VERSION


f = open(os.path.join(os.path.dirname(__file__), 'README.rst'))
readme = f.read()
f.close()

__plugin_name__ = 'Updatorr'
__version__ = '.'.join(map(str, VERSION))
__description__ = 'Deluge plugin for automatic torrents updates'
__long_description__ = readme
__author__ = "Igor 'idle sign' Starikov"
__author_email__ = 'idlesign@yandex.ru'
__license__ = 'BSD License'
__url__ = 'http://github.com/idlesign/deluge-updatorr'
__pkg_data__ = {__plugin_name__.lower(): ['tracker_handlers/*', 'data/*']}

setup(
    name=__plugin_name__,
    version=__version__,
    description=__description__,
    long_description=__long_description__,
    author=__author__,
    author_email=__author_email__,
    url=__url__,
    license=__license__,
    install_requires=['setuptools'],
    zip_safe=False,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ],

    packages=[__plugin_name__.lower()],
    package_data=__pkg_data__,

    entry_points='''
    [deluge.plugin.core]
    %s = %s:CorePlugin
    [deluge.plugin.gtkui]
    %s = %s:GtkUIPlugin
    [deluge.plugin.web]
    %s = %s:WebUIPlugin
    ''' % ((__plugin_name__, __plugin_name__.lower()) * 3)
)



import os
import logging
import pkg_resources
from deluge.plugins.pluginbase import WebPluginBase

log = logging.getLogger(__name__)


def get_resource(filename):
    return pkg_resources.resource_filename('updatorr', os.path.join('data', filename))


class WebUI(WebPluginBase):
    scripts = [get_resource('updatorr.js')]
    debug_scripts = scripts
